import json
import sys
import os
import socket
from datetime import datetime
from collections import namedtuple
from .utils import merge_dicts

LoggingLevel = namedtuple('LoggingLevel', ['name', 'level'])
PinoConfig = namedtuple('PinoConfig', [
    'level', 'stream', 'enabled', 'bindings', 'messagekey', 'millidiff', 'parent'
])

DEBUG = LoggingLevel("debug", 10)
INFO = LoggingLevel("info", 20)
WARN = LoggingLevel("warn", 30)
ERROR = LoggingLevel("error", 40)
CRITICAL = LoggingLevel("critical", 50)

LEVELS = [DEBUG, INFO, WARN, ERROR, CRITICAL]
LEVEL_NAMES = [level.name for level in LEVELS]
LEVEL_BY_NAME = {level.name: level for level in LEVELS}
LEVEL_BY_CODE = {level.level: level for level in LEVELS}

def get_level(level_name_or_code):
    if isinstance(level_name_or_code, LoggingLevel):
        return level_name_or_code
    if isinstance(level_name_or_code, int):
        return LEVEL_BY_CODE.get(level_name_or_code)
    return LEVEL_BY_NAME.get(level_name_or_code)


host = socket.gethostname()


def get_logger(self, level):
    metas = self._config.bindings or {}
    stream = self._config.stream
    message_key = self._config.messagekey
    should_millidiff = self._config.millidiff

    def log(*args, **kwargs):
        has_meta = isinstance(args[0], dict)

        if has_meta:
            message_metas = args[0]
            complete_metas = merge_dicts(message_metas, metas)
            args = args[1:] # shift args
        else:
            complete_metas = metas

        if len(args) > 1:
            message = args[0] % args[1:]
        elif len(kwargs):
            message = args[0].format(**kwargs)
        else:
            message = args[0]

        now = int(1000* datetime.now().timestamp())
        json_log = {
            "level": level,
            "time": now,
            message_key: message,
            "host": host,
            **complete_metas
        }
        if should_millidiff:
            delta = (now - self._last_timestamp) if self._last_timestamp else 0
            json_log["millidiff"] = delta
            self._last_timestamp = now
        stream.write(self._dumps(json_log))
        stream.write(os.linesep)
        stream.flush()
    log.__name__ = level
    return log

class DummyLogger:
    def critical(self, *args, **kwargs):
        pass
    def error(self, *args, **kwargs):
        pass
    def warn(self, *args, **kwargs):
        pass
    def info(self, *args, **kwargs):
        pass
    def debug(self, *args, **kwargs):
        pass

class PinoLogger(DummyLogger):
    __slots__ = ["_config", "_last_timestamp", "_dumps"]

    def __init__(
        self,
        bindings=None, level="info", stream=sys.stdout,
        enabled=True, parent=None, millidiff=True, messagekey="message",
        dump_function=json.dumps
    ):
        logging_level = get_level(level)
        self._config = PinoConfig(logging_level, stream, enabled, bindings, messagekey, millidiff, parent)
        self._last_timestamp = None
        self._setup_logging(self._config)
        self._dumps = dump_function

    def _setup_logging(self, config):
        if config.enabled:
            for level in LEVELS:
                logging_method = get_logger(self, level.name) if level.level >= config.level.level \
                    else getattr(super(), level.name)
                setattr(self, level.name, logging_method)

    @property
    def level(self):
        return self._config.level.name

    @level.setter
    def level(self, new_level):
        self._config = self._config._replace(level=get_level(new_level))
        self._setup_logging(self._config)

    def child(self, metas=None, **kwargs_metas):
        merged_bindings = merge_dicts(metas or kwargs_metas, self._config.bindings)
        child_logger = PinoLogger(
            **self._config._replace(parent=self, bindings=merged_bindings)._asdict(),
            dump_function=self._dumps
        )
        child_logger._last_timestamp = self._last_timestamp
        return child_logger




def merge_dicts(dict_a, dict_b):
    """Return dict issue from merge from dict_a and dict_b, dict_a having precedence"""
    # inspired from https://stackoverflow.com/questions/7204805/dictionaries-of-dictionaries-merge/7205107#7205107
    if not dict_a:
        return dict_b
    if not dict_b:
        return dict_a
    merged_dict = dict(dict_a)
    for key in dict_b:
        if key in dict_a:
            if isinstance(dict_a[key], dict) and isinstance(dict_b[key], dict):
                merged_dict[key] = merge_dicts(dict_a[key], dict_b[key])
        else:
            merged_dict[key] = dict_b[key]
    return merged_dict



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pino']

package_data = \
{'': ['*']}

install_requires = \
['style>=1.1.6,<2.0.0']

entry_points = \
{'console_scripts': ['pino-pretty = pino.pretty_cli:main']}

setup_kwargs = {
    'name': 'pino',
    'version': '0.6.0',
    'description': 'Python json logger inspired by pino.js',
    'long_description': 'pino.py\n=======\n\n> **Json natural logger for python** inspired by [pino.js](https://github.com/pinojs/pino) :evergreen_tree:\n\n[![PyPI](https://img.shields.io/pypi/v/pino.svg)](https://pypi.org/project/pino/)\n[![Supported Python Versions](https://img.shields.io/pypi/pyversions/pino.svg)](https://pypi.python.org/pypi/pino)\n[![Build Status](https://travis-ci.com/CoorpAcademy/pino.py.svg?branch=master)](https://travis-ci.com/CoorpAcademy/pino.py)\n[![codecov](https://codecov.io/gh/CoorpAcademy/pino.py/branch/master/graph/badge.svg)](https://codecov.io/gh/CoorpAcademy/pino.py)\n\n> In building port of [pinojs](https://github.com/pinojs/pino) logging library to python :snake:\n\n:warning: This is a in building prototype, it\'s API is subject to change.\nA CHANGELOG will be introduced once it\'s stable enough and publicized.\nUse it at you own risk, but feel free to reach with an issue.\n\n\n## Basic Example\n\n```python\nfrom pino import pino\n\nlogger = pino(\n    bindings={"apptype": "prototype", "context": "main"}\n)\n\nlogger.info("Hello, I just started")\nlogger.debug({"details": 42}, "Some details that won\'t be seen")\n\nchild_logger = logger.child(context="some_job")\nchild_logger.info("Job started")\nchild_logger.info({"duration": 4012}, "Job completed %s", "NOW")\n\nlogger.info("Program completed")\n```\n\nWhich would output:\n```\n{"level": "info", "time": 1587740056952, "message": "Hello, I just started", "host": "SomeHost", "apptype": "prototype", "context": "main", "millidiff": 0}\n{"level": "info", "time": 1587740056952, "message": "Job started", "host": "SomeHost", "context": "some_job", "apptype": "prototype", "millidiff": 0}\n{"level": "info", "time": 1587740056952, "message": "Job completed NOW", "host": "SomeHost", "duration": 4012, "context": "some_job", "apptype": "prototype", "millidiff": 0}\n{"level": "info", "time": 1587740056952, "message": "Program completed", "host": "SomeHost", "apptype": "prototype", "context": "main", "millidiff": 0}\n```\n\n## API\n### pino() constructor arguments\n\n- `bindings`: meta attached to the messages by default\n- `level`: minimal level to output logs, _default to `info`_\n- `enabled`: is logger enabled, _default to true_\n- `millidiff`: whether a millidiff is added to message, `ms` since last message, _enabled by default_.\n- `stream` : stream to write logs to, default to `sys.stdout`\n- `dump_function`: function to be used to serialise object to JSON, _default `json.dumps`_\n- `messagekey`: key for message entry,  _default `message`_\n\n### pino logger instance\n- log methods: `critical`, `error`, `warn`, `info`, `debug`: (extra_bindings?), message, template value\n- `.level`: access or update current log level\n\n- `child(metas)`: create a child logger instance with new metas/bindings attached to it. (metas can be provided either as dict or kwargs)\n\n### Complex examples\n\nYou can see more detailed examples in the [**examples** folder](./examples), notably [complex.py](./examples/complex.py)\n\n## Development :hammer_and_wrench:\n\nThis library use [***Poetry***](https://python-poetry.org/) for management of the project, it\'s dependencies, build and else.\n\nYou can run test on all supported python version with `poetry run task test` (which will run `tox`),\nor you can run on your current python version with `poetry run task pytest`.\n',
    'author': 'Adrien Becchis',
    'author_email': 'adrien.becchis@coorpacademy.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CoorpAcademy/pino.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
